@@ -9,24 +9,18 @@ import logging
 from typing import Any, Dict, List, Optional, Union, Callable
 from datetime import datetime, timedelta
 
+from .exceptions import (
+    MCPError, # Base MCPError can still be used for very generic cases if needed
+    MCPClientError,
+    MCPConnectionError,
+    MCPTimeoutError,
+    MCPToolError,
+    MCPAuthenticationError,
+    MCPResponseParseError
+)
+
 logger = logging.getLogger(__name__)
 
-class MCPError(Exception):
-    """Base error for MCP-related issues"""
-    pass
-
-class MCPConnectionError(MCPError):
-    """Error connecting to MCP server"""
-    pass
-
-class MCPTimeoutError(MCPError):
-    """Timeout during MCP request"""
-    pass
-
-class MCPToolError(MCPError):
-    """Error executing an MCP tool"""
-    pass
-
 class MCPTools:
     """Container for MCP tool information with cache management"""
     
@@ -155,9 +149,11 @@ class MCPClient:
             List of available tools with their metadata
             
         Raises:
-            MCPConnectionError: If connection to the server fails
-            MCPTimeoutError: If the request times out
-            MCPError: For other errors
+            MCPAuthenticationError: If authentication fails.
+            MCPConnectionError: If connection to the server fails.
+            MCPTimeoutError: If the request times out.
+            MCPResponseParseError: If the server response is unparseable.
+            MCPClientError: For other client-side MCP errors.
         """
         # Return cached tools if available and not stale
         if self._tools_cache is not None and not force_refresh:
@@ -171,39 +167,46 @@ class MCPClient:
         while retry_count <= self.max_retries:
             try:
                 response = await self.client.get(f"{self.server_url}/tools")
-                response.raise_for_status()
+                response.raise_for_status() # Raises HTTPStatusError for 4xx/5xx
                 
-                tools = response.json()
+                try:
+                    tools = response.json()
+                except json.JSONDecodeError as e:
+                    logger.error(f"Failed to parse JSON response from MCP server: {e}")
+                    raise MCPResponseParseError(f"Failed to parse JSON response from {self.server_url}/tools: {str(e)}")
+
                 self._tools_cache = MCPTools(tools, ttl=self.tools_ttl)
                 return tools
                 
             except httpx.TimeoutException as e:
                 retry_count += 1
                 if retry_count > self.max_retries:
-                    logger.error(f"Timeout getting tools from MCP server after {self.max_retries} retries")
-                    raise MCPTimeoutError(f"Timeout getting tools from MCP server: {str(e)}")
-                
-                logger.warning(f"Timeout getting tools from MCP server, retrying ({retry_count}/{self.max_retries})")
+                    logger.error(f"Timeout getting tools from MCP server {self.server_url} after {self.max_retries} retries")
+                    raise MCPTimeoutError(f"Timeout getting tools from MCP server {self.server_url}: {str(e)}")
+                logger.warning(f"Timeout getting tools from MCP server {self.server_url}, retrying ({retry_count}/{self.max_retries})")
                 await asyncio.sleep(delay)
                 delay *= 2  # Exponential backoff
                 
             except httpx.HTTPStatusError as e:
-                logger.error(f"HTTP error getting tools from MCP server: {e}")
-                raise MCPConnectionError(f"HTTP error getting tools from MCP server: {e.response.status_code} - {e.response.text}")
+                logger.error(f"HTTP error getting tools from MCP server {self.server_url}: {e.response.status_code} - {e.response.text}")
+                if e.response.status_code in (401, 403):
+                    raise MCPAuthenticationError(f"Authentication failed for {self.server_url}/tools: {e.response.status_code} - {e.response.text}")
+                raise MCPConnectionError(f"HTTP error getting tools from MCP server {self.server_url}: {e.response.status_code} - {e.response.text}")
                 
-            except httpx.RequestError as e:
+            except httpx.RequestError as e: # Covers connection errors, DNS issues etc.
                 retry_count += 1
                 if retry_count > self.max_retries:
-                    logger.error(f"Request error getting tools from MCP server after {self.max_retries} retries")
-                    raise MCPConnectionError(f"Request error getting tools from MCP server: {str(e)}")
-                
-                logger.warning(f"Request error getting tools from MCP server, retrying ({retry_count}/{self.max_retries})")
+                    logger.error(f"Request error getting tools from MCP server {self.server_url} after {self.max_retries} retries: {str(e)}")
+                    raise MCPConnectionError(f"Request error getting tools from MCP server {self.server_url}: {str(e)}")
+                logger.warning(f"Request error getting tools from MCP server {self.server_url}, retrying ({retry_count}/{self.max_retries})")
                 await asyncio.sleep(delay)
                 delay *= 2  # Exponential backoff
                 
+            except MCPClientError: # Re-raise specific MCPClientErrors if they somehow occur here
+                raise
             except Exception as e:
-                logger.error(f"Unexpected error getting tools from MCP server: {e}")
-                raise MCPError(f"Failed to get tools from MCP server: {str(e)}")
+                logger.error(f"Unexpected error getting tools from MCP server {self.server_url}: {e}")
+                raise MCPClientError(f"Failed to get tools from MCP server {self.server_url}: {str(e)}")
             
     async def call_tool(
         self, 
@@ -225,10 +228,12 @@ class MCPClient:
             Result from the tool
             
         Raises:
-            MCPConnectionError: If connection to the server fails
-            MCPTimeoutError: If the request times out
-            MCPToolError: If the tool execution fails
-            MCPError: For other errors
+            MCPAuthenticationError: If authentication fails.
+            MCPConnectionError: If connection to the server fails.
+            MCPTimeoutError: If the request times out.
+            MCPToolError: If the tool execution fails.
+            MCPResponseParseError: If the server response is unparseable.
+            MCPClientError: For other client-side MCP errors.
         """
         retry_count = 0
         delay = self.retry_delay
@@ -243,9 +248,13 @@ class MCPClient:
                     f"{self.server_url}/tools/{tool_name}",
                     json=params
                 )
-                response.raise_for_status()
+                response.raise_for_status() # Raises HTTPStatusError for 4xx/5xx
                 
-                result = response.json()
+                try:
+                    result = response.json()
+                except json.JSONDecodeError as e:
+                    logger.error(f"Failed to parse JSON response from MCP tool {tool_name} on {self.server_url}: {e}")
+                    raise MCPResponseParseError(f"Failed to parse JSON response for tool {tool_name} from {self.server_url}: {str(e)}")
                 
                 # Check for error in the response
                 if isinstance(result, dict) and result.get("isError", False):
@@ -272,34 +281,35 @@ class MCPClient:
             except httpx.TimeoutException as e:
                 retry_count += 1
                 if retry_count > self.max_retries:
-                    logger.error(f"Timeout calling MCP tool {tool_name} after {self.max_retries} retries")
-                    raise MCPTimeoutError(f"Timeout calling MCP tool {tool_name}: {str(e)}")
-                
-                logger.warning(f"Timeout calling MCP tool {tool_name}, retrying ({retry_count}/{self.max_retries})")
+                    logger.error(f"Timeout calling MCP tool {tool_name} on {self.server_url} after {self.max_retries} retries")
+                    raise MCPTimeoutError(f"Timeout calling MCP tool {tool_name} on {self.server_url}: {str(e)}")
+                logger.warning(f"Timeout calling MCP tool {tool_name} on {self.server_url}, retrying ({retry_count}/{self.max_retries})")
                 await asyncio.sleep(delay)
                 delay *= 2  # Exponential backoff
                 
             except httpx.HTTPStatusError as e:
-                logger.error(f"HTTP error calling MCP tool {tool_name}: {e}")
-                raise MCPToolError(f"HTTP error calling MCP tool {tool_name}: {e.response.status_code} - {e.response.text}")
+                logger.error(f"HTTP error calling MCP tool {tool_name} on {self.server_url}: {e.response.status_code} - {e.response.text}")
+                if e.response.status_code in (401, 403):
+                    raise MCPAuthenticationError(f"Authentication failed for tool {tool_name} on {self.server_url}: {e.response.status_code} - {e.response.text}")
+                # For other HTTP errors, it's often a tool-specific issue or server misconfiguration
+                raise MCPToolError(f"HTTP error calling MCP tool {tool_name} on {self.server_url}: {e.response.status_code} - {e.response.text}")
                 
-            except httpx.RequestError as e:
+            except httpx.RequestError as e: # Covers connection errors, DNS issues etc.
                 retry_count += 1
                 if retry_count > self.max_retries:
-                    logger.error(f"Request error calling MCP tool {tool_name} after {self.max_retries} retries")
-                    raise MCPConnectionError(f"Request error calling MCP tool {tool_name}: {str(e)}")
-                
-                logger.warning(f"Request error calling MCP tool {tool_name}, retrying ({retry_count}/{self.max_retries})")
+                    logger.error(f"Request error calling MCP tool {tool_name} on {self.server_url} after {self.max_retries} retries: {str(e)}")
+                    raise MCPConnectionError(f"Request error calling MCP tool {tool_name} on {self.server_url}: {str(e)}")
+                logger.warning(f"Request error calling MCP tool {tool_name} on {self.server_url}, retrying ({retry_count}/{self.max_retries})")
                 await asyncio.sleep(delay)
                 delay *= 2  # Exponential backoff
                 
-            except MCPToolError:
-                # Don't retry tool execution errors
+            except MCPToolError: # Re-raise specific MCPToolError if it's from the "isError" block
+                raise
+            except MCPClientError: # Re-raise other specific MCPClientErrors
                 raise
-                
             except Exception as e:
-                logger.error(f"Unexpected error calling MCP tool {tool_name}: {e}")
-                raise MCPError(f"Failed to call MCP tool {tool_name}: {str(e)}")
+                logger.error(f"Unexpected error calling MCP tool {tool_name} on {self.server_url}: {e}")
+                raise MCPClientError(f"Failed to call MCP tool {tool_name} on {self.server_url}: {str(e)}")
     
     async def _stream_tool_call(
         self, 
@@ -326,8 +336,20 @@ class MCPClient:
             f"{self.server_url}/tools/{tool_name}",
             json=params
         ) as response:
-            response.raise_for_status()
-            
+            try:
+                response.raise_for_status() # Check for HTTP errors before streaming
+            except httpx.HTTPStatusError as e:
+                logger.error(f"HTTP error streaming MCP tool {tool_name} on {self.server_url}: {e.response.status_code} - {e.response.text}")
+                if e.response.status_code in (401, 403):
+                    raise MCPAuthenticationError(f"Authentication failed for streaming tool {tool_name} on {self.server_url}: {e.response.status_code} - {e.response.text}")
+                raise MCPToolError(f"HTTP error streaming MCP tool {tool_name} on {self.server_url}: {e.response.status_code} - {e.response.text}")
+            except httpx.TimeoutException as e:
+                 logger.error(f"Timeout streaming MCP tool {tool_name} on {self.server_url}: {str(e)}")
+                 raise MCPTimeoutError(f"Timeout streaming MCP tool {tool_name} on {self.server_url}: {str(e)}")
+            except httpx.RequestError as e:
+                 logger.error(f"Request error streaming MCP tool {tool_name} on {self.server_url}: {str(e)}")
+                 raise MCPConnectionError(f"Request error streaming MCP tool {tool_name} on {self.server_url}: {str(e)}")
+
             async for chunk in response.aiter_bytes():
                 chunk_str = chunk.decode("utf-8")
                 full_response.append(chunk_str)
@@ -347,7 +369,8 @@ class MCPClient:
             Result from the tool
             
         Raises:
-            MCPError: For any errors
+            MCPClientError: For any client-side errors during the async call.
+            MCPError: For other MCP related errors.
         """
         loop = asyncio.get_event_loop()
         return loop.run_until_complete(self.call_tool(tool_name, **params))
@@ -360,7 +383,8 @@ class MCPClient:
             List of available tools with their metadata
             
         Raises:
-            MCPError: For any errors
+            MCPClientError: For any client-side errors during the async call.
+            MCPError: For other MCP related errors.
         """
         loop = asyncio.get_event_loop()
         return loop.run_until_complete(self.get_tools())
